/*
**  SDLSystem.m
**
**  Copyright (c) 2002, 2003
**
**  Author: Yen-Ju Chen <yjchenx@hotmail.com>
**          Part of the code is taken from playsound in SDL_sound
**
**  This program is free software; you can redistribute it and/or modify
**  it under the terms of the GNU Lesser General Public License as published by
**  the Free Software Foundation; either version 2 of the License, or
**  (at your option) any later version.
**
**  This program is distributed in the hope that it will be useful,
**  but WITHOUT ANY WARRANTY; without even the implied warranty of
**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**  GNU General Public License for more details.
**
**  You should have received a copy of the GNU Lesser General Public License
**  along with this program; if not, write to the Free Software
**  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "SDLSystem.h"
#include <SDL.h>
#include <SDL_thread.h>
#include <SDL/SDL_sound.h>
#include <string.h>
#include <Foundation/NSArray.h>
#include <Foundation/NSString.h>
#include <Foundation/NSValue.h>
#include <Foundation/NSDictionary.h>

#define DEFAULT_DECODEBUF 16384
#define DEFAULT_AUDIOBUF  4096

struct SDLSystem_t {
  @defs(SDLSystem)
} *SDLSystemPtr;

/* global decoding state. */
typedef struct
{
    Uint8 *decoded_ptr;
    Uint32 decoded_bytes;
    Uint32 total_bytes;
    int wants_volume_change;
    float volume;
    Uint32 seek_list;
} playsound_global_state;

static volatile playsound_global_state global_state;
static int currentTime;

static BOOL isPlaying;
static BOOL isPaused;
static BOOL hasCDROM;

static enum player_control controller = PLAYER_CONTROL_STOP;
static SDLSystem *instance = nil;
static SDL_CD *cdrom = NULL;

@implementation SDLSystem

static Uint32 cvtMsToBytePos(Sound_AudioInfo *info, Uint32 ms)
{
    /* "frames" == "sample frames" */
    float frames_per_ms = ((float) info->rate) / 1000.0;
    Uint32 frame_offset = (Uint32) (frames_per_ms * ((float) ms));
    Uint32 frame_size = (Uint32) ((info->format & 0xFF) / 8) * info->channels;
    return(frame_offset * frame_size);
} /* cvtMsToBytePos */

static Uint32 cvtByteToS(Sound_AudioInfo *info, Uint32 len)
{
   float frames_per_ms = ((float) info->rate);
   Uint32 frame_size = (Uint32) ((info->format & 0xFF) / 8) * info->channels;
   Uint32 frame_offset = len / frame_size;
   return (frame_offset/frames_per_ms);
}

static int read_more_data(Sound_Sample *sample)
{
    if (controller == PLAYER_CONTROL_STOP)  /* probably a sigint; stop trying to read. */
    {
        global_state.decoded_bytes = 0;
        return(0);
    } /* if */

    if (global_state.decoded_bytes > 0) /* don't need more data; just return. */
        return(global_state.decoded_bytes);

      /* Need more audio data. Check seek... */
    if (global_state.seek_list != -1)
    {
        Uint32 seek_list = global_state.seek_list;

        NSLog(@"Seeking to %.2d:%.2d:%.4d...",
                (int) ((seek_list / 1000) / 60),
                (int) ((seek_list / 1000) % 60),
                (int) ((seek_list % 1000)));

        if(!Sound_Seek(sample, seek_list))
          {
            NSLog(@"Sound_Seek() failed");
            controller = PLAYER_CONTROL_STOP;
          } /* if */
        global_state.total_bytes = cvtMsToBytePos(&sample->desired,
                                                  global_state.seek_list);
        global_state.seek_list = -1; /* Finish seek */

        return(read_more_data(sample));  /* handle loops conditions. */
    } /* if */

        /* See if there's more to be read... */
    if (!(sample->flags & (SOUND_SAMPLEFLAG_ERROR | SOUND_SAMPLEFLAG_EOF)) )
    {
//        NSLog(@"Sound_Decode");
        global_state.decoded_bytes = Sound_Decode(sample);
        if (sample->flags & SOUND_SAMPLEFLAG_ERROR)
        {
            NSLog(@"Error in decoding sound file!");
        } /* if */

        global_state.decoded_ptr = sample->buffer;
        return(read_more_data(sample));  /* handle loops conditions. */
    } /* if */

    /* No more to be read from stream, but we may want to loop the sample. */
    return(0);
} /* read_more_data */

static void memcpy_with_volume(Sound_Sample *sample,
                               Uint8 *dst, Uint8 *src, int len)
{
    int i;
    Uint16 *u16src = NULL;
    Uint16 *u16dst = NULL;
    Sint16 *s16src = NULL;
    Sint16 *s16dst = NULL;
    float volume = global_state.volume;

    if (!global_state.wants_volume_change)
    {
        memcpy(dst, src, len);
        return;
    } /* if */

    /* !!! FIXME: This would be more efficient with a lookup table. */
    switch (sample->desired.format)
    {
        case AUDIO_U8:
            for (i = 0; i < len; i++, src++, dst++)
                *dst = (Uint8) (((float) (*src)) * volume);
            break;

        case AUDIO_S8:
            for (i = 0; i < len; i++, src++, dst++)
                *dst = (Sint8) (((float) (*src)) * volume);
            break;

        case AUDIO_U16LSB:
            u16src = (Uint16 *) src;
            u16dst = (Uint16 *) dst;
            for (i = 0; i < len; i += sizeof (Uint16), u16src++, u16dst++)
            {
                *u16dst = (Uint16) (((float) (SDL_SwapLE16(*u16src))) * volume);
                *u16dst = SDL_SwapLE16(*u16dst);
            } /* for */
            break;

        case AUDIO_S16LSB:
            s16src = (Sint16 *) src;
            s16dst = (Sint16 *) dst;
            for (i = 0; i < len; i += sizeof (Sint16), s16src++, s16dst++)
            {
                *s16dst = (Sint16) (((float) (SDL_SwapLE16(*s16src))) * volume);
                *s16dst = SDL_SwapLE16(*s16dst);
            } /* for */
            break;

        case AUDIO_U16MSB:
            u16src = (Uint16 *) src;
            u16dst = (Uint16 *) dst;
            for (i = 0; i < len; i += sizeof (Uint16), u16src++, u16dst++)
            {
                *u16dst = (Uint16) (((float) (SDL_SwapBE16(*u16src))) * volume);
                *u16dst = SDL_SwapBE16(*u16dst);
            } /* for */
            break;

        case AUDIO_S16MSB:
            s16src = (Sint16 *) src;
            s16dst = (Sint16 *) dst;
            for (i = 0; i < len; i += sizeof (Sint16), s16src++, s16dst++)
            {
                *s16dst = (Sint16) (((float) (SDL_SwapBE16(*s16src))) * volume);
                *s16dst = SDL_SwapBE16(*s16dst);
            } /* for */
            break;
    } /* switch */
} /* memcpy_with_volume */

static void audio_callback(void *userdata, Uint8 *stream, int len)
{
    Sound_Sample *sample = (Sound_Sample *) userdata;
    int bw = 0; /* bytes written to stream this time through the callback */

    while (bw < len)
    {
        int cpysize;  /* bytes to copy on this iteration of the loop. */

        if (!read_more_data(sample)) /* read more data, if needed. */
        {
            /* ...there isn't any more data to read! */
            memset(stream + bw, '\0', len - bw);
            controller = PLAYER_CONTROL_STOP;
            return;
        } /* if */

        /* decoded_bytes and decoder_ptr are updated as necessary... */

        cpysize = len - bw;
        if (cpysize > global_state.decoded_bytes)
            cpysize = global_state.decoded_bytes;

        if (cpysize > 0)
        {
            memcpy_with_volume(sample, stream + bw,
                               (Uint8 *) global_state.decoded_ptr,
                               cpysize);

            bw += cpysize;
            global_state.decoded_ptr += cpysize;
            global_state.decoded_bytes -= cpysize;
            global_state.total_bytes += cpysize;
            currentTime = cvtByteToS(&sample->desired, global_state.total_bytes);
        } /* if */
    } /* while */
} /* audio_callback */

int thread_play(void *data)
{
  int delay;
  Uint32 decode_buffersize;
  Uint32 audio_buffersize;
  SDL_RWops *rwops;
  Sound_Sample *sample;
  SDL_AudioSpec sdl_desired;

  memset((void *) &global_state, '\0', sizeof (global_state));
  global_state.seek_list = -1;
  global_state.volume = 1.0;
  global_state.total_bytes = 0;
  audio_buffersize = DEFAULT_AUDIOBUF;
  decode_buffersize = DEFAULT_DECODEBUF;

//  rwops = SDL_RWFromFile(filename, "r");
  rwops = SDL_RWFromFile([SDLSystemPtr->filePath cString], "r");

  sample = Sound_NewSample(rwops, 
                           [[SDLSystemPtr->filePath pathExtension] cString],
                           NULL,
                           decode_buffersize);
/*
  sample = Sound_NewSampleFromFile(filename,
                                   NULL,
                                   decode_buffersize);
*/

  if (sample == NULL)
    {
      NSLog(@"Can't load file: %@ (thread_Play)", SDLSystemPtr->filePath);
      return NO;
    }

  sdl_desired.freq = sample->actual.rate;
  sdl_desired.format = sample->actual.format;
  sdl_desired.channels = sample->actual.channels;
  sdl_desired.samples = audio_buffersize;
  sdl_desired.callback = audio_callback;
  sdl_desired.userdata = sample;

  NSLog(@"freq (rate), %d, channels %d", sdl_desired.freq,
                                         sdl_desired.channels);


  if (SDL_OpenAudio(&sdl_desired, NULL) < 0)
    {
       NSLog(@"Couldn't open audio device!");
       Sound_Quit();
       SDL_Quit();
       return NO;
    }
  SDL_PauseAudio(0);

  controller = PLAYER_CONTROL_REPLAY;

  while (1)
    {
      if (controller == PLAYER_CONTROL_REPLAY) 
        {
        }
      else if (controller == PLAYER_CONTROL_STOP)
        {
           controller = PLAYER_CONTROL_REPLAY;
           break;
        }
      else if (controller == PLAYER_CONTROL_PAUSE)
        {
           SDL_PauseAudio(1);
           controller = PLAYER_CONTROL_REPLAY;
        }
      else if (controller == PLAYER_CONTROL_RESUME)
        {
           SDL_PauseAudio(0);
           controller = PLAYER_CONTROL_REPLAY;
        }
      SDL_Delay(10);
    }

  SDL_PauseAudio(1);

  /*
   * Sleep two buffers' worth of audio before closing, in order
   *  to allow the playback to finish. This isn't always enough;
   *   perhaps SDL needs a way to explicitly wait for device drain?
   */
  /* Since two is not enough, use ten instead */
  delay = 4 * 1000 * sdl_desired.samples / sdl_desired.freq;
  SDL_Delay(delay);

  SDL_CloseAudio();  /* reopen with next sample's format if possible */

  Sound_FreeSample(sample);
//  SDL_RWclose(rwops);
  NSLog(@"SDL_RWclose");
//  SDL_FreeRW(rwops);
  NSLog(@"SDL_FreeRW");

  global_state.seek_list = -1;
  global_state.volume = 1.0;
  global_state.total_bytes = 0;

  isPlaying = NO;
  isPaused = NO;

  if (SDLSystemPtr->_delegate)
    {
      [SDLSystemPtr->_delegate soundFinishPlaying];
    }

  return 0;
}

- (BOOL) seekSound: (int) seconds
{
  NSLog(@"SDLWrapper seekSound %d", seconds); 
  global_state.seek_list = seconds * 1000;
  return YES;
}

- (BOOL) playSound: (NSString *) aFile
{
  filePath = [aFile stringByExpandingTildeInPath];

  controller = PLAYER_CONTROL_REPLAY;

  if (isPlaying == NO)
    {
      SDL_Thread *thread;
 
      // Start playing
      isPlaying = YES;

      thread = SDL_CreateThread(thread_play, (void *)NULL);
    }
  return YES;
}

- (NSArray *) availableFormats
{
  const Sound_DecoderInfo **rc = Sound_AvailableDecoders();
  const Sound_DecoderInfo **i;
  const char **ext;
  NSString *extension;
  NSMutableArray *array = [NSMutableArray new];

  if (rc != NULL)
  {
    for (i = rc; *i != NULL; i++)
    {
      for (ext = (*i)->extensions; *ext != NULL; ext++) {
        extension = [[NSString stringWithFormat: @"%s", *ext]
                                          lowercaseString];
        if([array containsObject: extension] == NO) {
          [array addObject: extension];
        }
      }
    } 
  }
  return AUTORELEASE(array);
}

- (int) currentTimeOfSound
{
  return currentTime;
}

// Return length in seconds
- (int) lengthOfSound: (NSString *) filename
{
  Uint32 decode_buffersize;
  SDL_RWops *rwops;
  Sound_Sample *sample;
  int len, seconds;

  decode_buffersize = DEFAULT_DECODEBUF*10;

  rwops = SDL_RWFromFile([filename cString], "r");

  sample = Sound_NewSample(rwops,
                           [[filename pathExtension] cString],
                           NULL,
                           decode_buffersize);

  if (sample == NULL)
    {
      NSLog(@"Can't load file: %@ (lengthOfSound)", filename);
      return NO;
    }

  len = Sound_DecodeAll(sample);
  seconds = cvtByteToS(&sample->desired, len);

  Sound_FreeSample(sample);

  return seconds;
}

- (BOOL) hasCDROM
{
  return hasCDROM;
}

- (int) lengthOfCDTrack: (int) track
{
  BOOL needToBeClose = NO;
  int m = 0, s = 0, f = 0;

  if (hasCDROM == NO)
    return 0;

  if (!cdrom)
    {
      [self openCD];
      needToBeClose = YES;
    }

  if(CD_INDRIVE(SDL_CDStatus(cdrom))) 
    {
      FRAMES_TO_MSF(cdrom->track[track].length, &m, &s, &f);
    }

  if (needToBeClose == YES)
    {
      [self closeCD];
    }

  return (m*60+s);
}

- (int) currentTimeOfCDTrack
{
  if (hasCDROM == NO)
    return 0;

  if(CD_INDRIVE(SDL_CDStatus(cdrom)))
    return (cdrom->cur_frame/CD_FPS);
  return 0;
}

- (BOOL) playCD: (int) num
{
  if (hasCDROM == NO)
    return NO;

  if (!cdrom)
    [self openCD];

  if(CD_INDRIVE(SDL_CDStatus(cdrom)))
    if(!SDL_CDPlayTracks(cdrom, num, 0, 0, 0));
      return YES;
  NSLog(@"SDL can't play CD: track #%d", num);
  return NO;
}

- (BOOL) playCD: (int) num at: (int) seconds
{
  if (hasCDROM == NO)
    return NO;

  if (!cdrom)
    [self openCD];

  if(CD_INDRIVE(SDL_CDStatus(cdrom)))
    if(!SDL_CDPlayTracks(cdrom, num, seconds*CD_FPS, 0, 0));
      return YES;
  NSLog(@"SDL can't play CD: track #%d at %d seconds", num, seconds);
  return NO;

}

- (BOOL) stopCD
{
  if (hasCDROM == NO)
    return NO;

  if(CD_INDRIVE(SDL_CDStatus(cdrom)))
    if(!SDL_CDStop(cdrom)) {
      [self closeCD];
      return YES;
    }
  return NO;
}

- (BOOL) pauseCD
{
  if (hasCDROM == NO)
    return NO;

  if(CD_INDRIVE(SDL_CDStatus(cdrom)))
    if(!SDL_CDPause(cdrom))
      return YES;
  return NO;
}

- (BOOL) resumeCD
{
  if (hasCDROM == NO)
    return NO;

  if(CD_INDRIVE(SDL_CDStatus(cdrom)))
    if(!SDL_CDResume(cdrom))
      return YES;
  return NO;
}

- (BOOL) isCDPlaying
{
  CDstatus status;

  if (hasCDROM == NO)
    return NO;

  status = SDL_CDStatus(cdrom);
  if (status == CD_PLAYING)
    return YES;
  return NO;
}

- (int) numberOfCDTracks
{
  if (hasCDROM == NO)
    return 0;

  return  cdrom->numtracks;
}

- (BOOL) isCDPresent
{
  if (hasCDROM == NO)
    return NO;

  if(CD_INDRIVE(SDL_CDStatus(cdrom)))
    return YES;
  else
    return NO;
}

- (BOOL) ejectCD
{
  if (hasCDROM == NO)
    return NO;

  if (!cdrom)
    [self openCD];

  if(SDL_CDEject(cdrom)) {
    NSLog(@"Failed to eject CD-ROM");
    return NO;
  }

  [self closeCD];

  return YES;
}

- (BOOL) openCD
{
  if (hasCDROM == NO)
    return NO;

  if (cdrom) {
    return YES;
  }

  cdrom = SDL_CDOpen(0);

  if (!cdrom)
    {
      NSLog(@"Can't open CD-ROM");
      return NO;
    }

  return YES;
}

- (void) closeCD
{
  if (hasCDROM && cdrom) {
    SDL_CDClose(cdrom);
    cdrom = NULL;
  }
}

/* Return array of CDTrack, must open CD first */
- (NSArray *) allCDTracks
{
  NSMutableDictionary *track;
  NSMutableArray *array = [NSMutableArray new];
  int i;

  if (hasCDROM == NO)
    return AUTORELEASE(array);

  if (!cdrom)
    [self openCD];

  if(CD_INDRIVE(SDL_CDStatus(cdrom)))
    {
      for (i = 0; i < cdrom->numtracks; i++)
        {
          track = [NSMutableDictionary new];
          [track setObject: [NSNumber numberWithInt: cdrom->track[i].length]
                    forKey: @"length"];
          [track setObject: [NSNumber numberWithInt: cdrom->track[i].offset]
                    forKey: @"offset"];
          [array addObject: track];
          RELEASE(track);
        }
    }
  [self closeCD];
  return AUTORELEASE(array);
}

- (id) init
{
  Uint32 sdl_init_flags = SDL_INIT_AUDIO | SDL_INIT_CDROM;

  self = [super init];

  cdrom = NULL;

  if (SDL_Init(sdl_init_flags) == -1)
    {
       NSLog(@"SDL init fails");
       return nil;
    }

  NSLog(@"CD-Rom drives %d", SDL_CDNumDrives());

  if(!SDL_CDNumDrives()) 
    {
       NSLog(@"No CD-ROM devices available");
       hasCDROM = NO;
//       SDL_Quit();
//       return nil;
    }
  else
    {
      hasCDROM = YES;
    }

  if (hasCDROM) {
    if ([self openCD] == NO)
      {
         SDL_Quit();
         return nil;
      }
    else
      [self closeCD];
  }

  if (!Sound_Init())
    {
       NSLog(@"SDL_sound init fails");   
       SDL_Quit();
       return nil;
    }

  isPlaying = NO;
  isPaused = NO;

  SDLSystemPtr = (struct SDLSystem_t *)self;

  return self;
}

+ (id) singleInstance
{
  if ( !instance)
    {
      NSLog(@"Load SDL system");
      instance = [[SDLSystem alloc] init];
    }
  return instance;
}

- (BOOL) isSoundPlaying
{
  return isPlaying;
}

- (BOOL) isSoundPaused
{
    return isPaused;
}    

- (void) setController: (enum player_control) aControl
{
  controller = aControl;
}

- (BOOL) resumeSound
{
  if (isPaused == YES)
    {
      [self setController: PLAYER_CONTROL_RESUME];
      isPaused = NO;
    }
  return YES;
}

- (BOOL) stopSound
{
  [self setController: PLAYER_CONTROL_STOP];
  isPlaying = NO;
  isPaused = NO;
  return YES;
}

- (BOOL) pauseSound
{
  if (isPlaying == YES)
    {
       [self setController: PLAYER_CONTROL_PAUSE];
       isPaused = YES;
    }
  return YES;
}

- (id) delegate
{
  return _delegate;
}

- (void) setDelegate: (id) aDelegate
{
  _delegate = aDelegate;
}

- (void) dealloc
{
  TEST_RELEASE(filePath);

  SDL_CDClose(cdrom);

  if (!Sound_Quit()) {
    NSLog(@"Failed to quit SDL_Sound");
  }

  SDL_Quit();

  [super dealloc];
}

@end
